package org.processmining.acceptingpetrinetreductionlayouter.algorithms;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinetreductionlayouter.models.Reduction;
import org.processmining.acceptingpetrinetreductionlayouter.parameters.LayoutParameters;
import org.processmining.acceptingpetrinetreductionlayouter.parameters.NamedColor;
import org.processmining.acceptingpetrinetreductionlayouter.workspaces.ReductionWorkspace;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.util.Pair;
import org.processmining.models.connections.GraphLayoutConnection;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

public class LayoutAlgorithm {

	private Collection<FinderAlgorithm> reductionAlgorithms;
	private ReductionWorkspace ws;
	private Reduction bestReduction;
	private double bestScore;

	public LayoutAlgorithm() {
		reductionAlgorithms = new ArrayList<FinderAlgorithm>();
		reductionAlgorithms.add(new AbstractionFinderAlgorithm());
		reductionAlgorithms.add(new TransitionFinderAlgorithm(false));
		reductionAlgorithms.add(new TransitionFinderAlgorithm(true));
		reductionAlgorithms.add(new PlaceFinderAlgorithm());
	}

	public AcceptingPetriNet apply(PluginContext context, AcceptingPetriNet net, LayoutParameters parameters) {
		ws = new ReductionWorkspace(net, parameters);
		try {
			ws.layoutConnection = context.getConnectionManager().getFirstConnection(GraphLayoutConnection.class, context, ws.net.getNet());
		} catch (ConnectionCannotBeObtained e) {
			ws.layoutConnection = new GraphLayoutConnection(ws.net.getNet());
		}
		
		GraphLayoutConnection layoutConnection = null;
		try {
			layoutConnection = context.getConnectionManager().getFirstConnection(GraphLayoutConnection.class, context, net.getNet());
		} catch (ConnectionCannotBeObtained e) {
			// Never mind.
		}
		for (Transition transition : ws.workNet.getNet().getTransitions()) {
			ws.layoutConnection.setSize(transition, new Dimension(1, 1));
		}
		if (layoutConnection != null) {
			for (Place place : ws.workNet.getNet().getPlaces()) {
				ws.layoutConnection.setSize(place, layoutConnection.getSize(ws.reversePlaceMap.get(place)));
			}
		}

		Map<Pair<Place, Place>, Transition> map = new HashMap<Pair<Place, Place>, Transition>();
		for (Transition transition : ws.workNet.getNet().getTransitions()) {
			Place sourcePlace = (Place) ws.workNet.getNet().getInEdges(transition).iterator().next().getSource();
			Place targetPlace = (Place) ws.workNet.getNet().getOutEdges(transition).iterator().next().getTarget();
			map.put(new Pair<Place, Place>(sourcePlace, targetPlace), transition);
		}

		while (reduce(ws))
			;
		layout(ws);
		ws.threshold = 0;
		while (sort(ws)) {
			layout(ws);
			ws.threshold += ws.parameters.getVerticalSpace() * (ws.SCALE / 10.0);
		}
		for (PetrinetNode node : ws.layout.keySet()) {
			System.out.println("[LayoutAlgorithm] " + node + ": " + ws.layout.get(node));
		}

		for (PetrinetEdge<?, ?> edge : ws.net.getNet().getEdges()) {
			ws.layoutConnection.setEdgePoints(edge, new ArrayList<Point2D>());
		}
		
		/*
		 * Here we start adapting the layout.
		 */
		for (Place place : ws.net.getNet().getPlaces()) {
			if (ws.placeMap.containsKey(place) && ws.layout.containsKey(ws.placeMap.get(place))) {
				ws.layoutConnection.setPosition(place,
						new Point2D.Double(ws.OFFSET + (ws.layout.get(ws.placeMap.get(place)).getFirst() / ws.SCALE),
								ws.OFFSET + (ws.layout.get(ws.placeMap.get(place)).getSecond() / ws.SCALE)));
			}
		}
		for (Transition transition : ws.net.getNet().getTransitions()) {
			if (ws.transitionMap.containsKey(transition) && ws.layout.containsKey(ws.transitionMap.get(transition))) {
				ws.layoutConnection.setPosition(transition,
						new Point2D.Double(
								ws.OFFSET + (ws.layout.get(ws.transitionMap.get(transition)).getFirst() / ws.SCALE),
								ws.OFFSET + (ws.layout.get(ws.transitionMap.get(transition)).getSecond() / ws.SCALE)));
			}
		}
		for (PetrinetEdge<?, ?> edge : ws.net.getNet().getEdges()) {
			List<Point2D> edgePoints = new ArrayList<Point2D>();
			if (edge instanceof Arc) {
				Arc arc = (Arc) edge;
				PetrinetNode sourceNode = arc.getSource();
				PetrinetNode targetNode = arc.getTarget();
				Place sourcePlace;
				Place targetPlace;
				if (sourceNode instanceof Place) {
					sourcePlace = ws.placeMap.get(sourceNode);
					targetPlace = ws.transitionMap.get(targetNode);
				} else {
					sourcePlace = ws.transitionMap.get(sourceNode);
					targetPlace = ws.placeMap.get(targetNode);
				}
				Pair<Double, Double> sourcePosition = ws.layout.get(sourcePlace);
				Pair<Double, Double> targetPosition = ws.layout.get(targetPlace);
				Dimension sourceDimension = ws.layoutConnection.getSize(sourceNode);
				Dimension targetDimension = ws.layoutConnection.getSize(targetNode);
				Transition transition = map.get(new Pair<Place, Place>(sourcePlace, targetPlace));
				if (parameters.getNamedColor() != NamedColor.ASIS && ws.hardTransitions.contains(transition)) {
					ws.reverseTransitionMap.get(transition).getAttributeMap().put(AttributeMap.EDGECOLOR, parameters.getNamedColor().getColor());
				}
				Pair<Double, Double> transitionPosition = ws.layout.get(transition);
				Dimension transitionDimension = new Dimension(0, 0);
				addEdges(edgePoints, getPoint(sourcePosition, sourceDimension),
						getPoint(targetPosition, targetDimension), getPoint(transitionPosition, transitionDimension),
						sourceDimension.getWidth() / 2.0 + 10.0, targetDimension.getWidth() / 2.0 + 10.0);
			}
			ws.layoutConnection.setEdgePoints(edge, edgePoints);
		}
//		return ws.workNet;
		ws.layoutConnection.setLayedOut(true);
		context.getConnectionManager().addConnection(ws.layoutConnection);
		return ws.net;
	}
	
	private void addEdges(List<Point2D> edgePoints, Point2D pointA, Point2D pointC, Point2D pointB, double distA,
			double distC) {
		if ((pointA.getX() > pointB.getX() && pointB.getX() > pointC.getX())
				|| (pointA.getX() < pointB.getX() && pointB.getX() < pointC.getX())) {
			//			if (Math.abs(pointA.getX() - pointC.getX()) > 80.0) {
			Point2D midPoint = getMidPoint(getFracPoint(pointA, pointB, 0.1), getFracPoint(pointC, pointB, 0.1));
			Point2D point = getFracPoint(pointA, midPoint, 0.25);
			if (distance(pointA, point) > distA && distance(pointC, point) > distC) {
				edgePoints.add(point);
			}
			point = getFracPoint(pointC, midPoint, 0.25);
			if (distance(pointA, point) > distA && distance(pointC, point) > distC) {
				edgePoints.add(point);
			}
			//			}
		} else {
			//			if (Math.abs(pointA.getX() - pointC.getX()) > 80.0) {
			
			Point2D midPoint = getMidPoint(getFracPoint(pointA, pointB, 0.0, 1.1), getFracPoint(pointC, pointB, 0.0, 1.1));
//			Point2D midPoint = getMidPoint(getFracPoint(pointA, pointB, -0.1, 1.1), getFracPoint(pointC, pointB, -0.1, 1.1));
			Point2D point = midPoint;
			if (distance(pointA, point) > distA && distance(pointC, point) > distC) {
				edgePoints.add(point);
			}
			//			}
		}
	}

	private double distance(Point2D pointA, Point2D pointB) {
		return Math.sqrt(Math.pow(pointA.getX() - pointB.getX(), 2.0) + Math.pow(pointA.getY() - pointB.getY(), 2.0));
	}

	private Point2D getPoint(Pair<Double, Double> position, Dimension size) {
		return new Point2D.Double(ws.OFFSET + (position.getFirst() / ws.SCALE) + 0.5 * size.getWidth(),
				ws.OFFSET + (position.getSecond() / ws.SCALE) + 0.5 * size.getHeight());
	}

	private Point2D getFracPoint(Point2D pointA, Point2D pointB, double frac) {
		return new Point2D.Double(frac * pointA.getX() + (1.0 - frac) * pointB.getX(),
				frac * pointA.getY() + (1.0 - frac) * pointB.getY());
	}

	private Point2D getFracPoint(Point2D pointA, Point2D pointB, double fracX, double fracY) {
		return new Point2D.Double(fracX * pointA.getX() + (1.0 - fracX) * pointB.getX(),
				fracY * pointA.getY() + (1.0 - fracY) * pointB.getY());
	}

	private Point2D getMidPoint(Point2D pointA, Point2D pointB) {
		return getFracPoint(pointA, pointB, 0.5);
	}

	public boolean reduce(ReductionWorkspace ws) {
		this.ws = ws;
		find();
		if (bestReduction != null) {
			System.out.println("[LayoutAlgorithm] " + bestScore + " " + bestReduction);
			bestReduction.apply(ws);
			return true;
		}
		return false;
	}

	private void find() {
		bestReduction = null;
		bestScore = -1.0;
		for (FinderAlgorithm reductionAlgorithm : reductionAlgorithms) {
			Collection<Reduction> reductions = reductionAlgorithm.apply(ws);
			for (Reduction reduction : reductions) {
				double score = reduction.score(ws);
				if (bestScore < score) {
					bestScore = score;
					bestReduction = reduction;
					if (reduction.hasOptimalScore(ws)) {
						return;
					}
				}
			}
		}
	}
	
	private void layout(ReductionWorkspace ws) {
		ws.layout = new HashMap<PetrinetNode,Pair<Double,Double>>();
		double y = 0.0;
		for (Place place : ws.workNet.getNet().getPlaces()) {
			System.out.println("[LayoutAlgorithm] Starting layout on place " + place.getLabel());
			while (ws.layoutMap.get(place).doSortLoops(false))
				;
			ws.layoutMap.get(place).doLayout(ws, new Pair<Double,Double>(0.0, y), false, ws.parameters.getHorizontalSpace() * ws.SCALE);
//			ws.layoutMap.get(place).doPrint(0);
			int[] height = ws.layoutMap.get(place).height();
			y += height[0] + height[1] + (ws.parameters.getVerticalSpace() * ws.SCALE);
		}
		for (Transition transition : ws.workNet.getNet().getTransitions()) {
			System.out.println("[LayoutAlgorithm] Starting layout on transition " + transition.getLabel());
			while (ws.layoutMap.get(transition).doSortLoops(false))
				;
			ws.layoutMap.get(transition).doLayout(ws, new Pair<Double,Double>(0.0, 0.0), false, ws.parameters.getHorizontalSpace() * ws.SCALE);
//			ws.layoutMap.get(transition).doPrint(0);
			int[] height = ws.layoutMap.get(transition).height();
			y += height[0] + height[1] + (ws.parameters.getVerticalSpace() * ws.SCALE);
		}
	}
	
	private boolean sort(ReductionWorkspace ws) {
		for (Place place : ws.workNet.getNet().getPlaces()) {
			if (ws.layoutMap.get(place).doSort(ws)) {
				return true;
			}
		}
		for (Transition transition : ws.workNet.getNet().getTransitions()) {
			if (ws.layoutMap.get(transition).doSort(ws)) {
				return true;
			}
		}
		return false;
	}
}
