package org.processmining.acceptingpetrinetreductionlayouter.models;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.processmining.acceptingpetrinetreductionlayouter.utils.NodeComparator;
import org.processmining.acceptingpetrinetreductionlayouter.workspaces.ReductionWorkspace;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Arc;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.semantics.petrinet.Marking;

public class AbstractionReduction implements Reduction {

	/*
	 * The place reduced.
	 */
	private Place place;
	/*
	 * The transition reduced.
	 */
	private Transition transition;
	/*
	 * The postset of the transition reduced.
	 */
	private List<Place> places;
	/*
	 * The preset of the place reduced.
	 */
	private List<Transition> transitions;

	public Place getPlace() {
		return place;
	}

	public void setPlace(Place place) {
		this.place = place;
	}

	public Transition getTransition() {
		return transition;
	}

	public void setTransition(Transition transition) {
		this.transition = transition;
	}

	public List<Place> getPlaces() {
		return places;
	}

	public void setPlaces(Collection<Place> places) {
		this.places = new ArrayList<Place>(places);
		Collections.sort(this.places, nodeComparator);
	}

	public List<Transition> getTransitions() {
		return transitions;
	}

	public void setTransitions(Collection<Transition> transitions) {
		this.transitions = new ArrayList<Transition>(transitions);
		Collections.sort(this.transitions, nodeComparator);
	}

	private NodeComparator nodeComparator = new NodeComparator();

	public void apply(ReductionWorkspace ws) {
		Set<Arc> arcsToReduce = new HashSet<Arc>();
		for (PetrinetEdge<?, ?> edge : ws.workNet.getNet().getEdges()) {
			if (edge instanceof Arc) {
				Arc arc = (Arc) edge;
				if (arc.getSource() == place || arc.getTarget() == place) {
					arcsToReduce.add(arc);
				}
				if (arc.getSource() == transition || arc.getTarget() == transition) {
					arcsToReduce.add(arc);
				}
			}
		}
		for (Arc arc : arcsToReduce) {
			ws.workNet.getNet().removeEdge(arc);
		}
		ws.workNet.getNet().removePlace(place);
		ws.workNet.getNet().removeTransition(transition);
		for (Transition transition : transitions) {
			for (Place place : places) {
				ws.workNet.getNet().addArc(transition, place);
			}
		}
		Marking initialMarking = new Marking(ws.workNet.getInitialMarking());
		drain(initialMarking, place);
		ws.workNet.setInitialMarking(initialMarking);
		Set<Marking> finalMarkings = new HashSet<Marking>();
		for (Marking marking : ws.workNet.getFinalMarkings()) {
			Marking finalMarking = new Marking(marking);
			drain(finalMarking, place);
			finalMarkings.add(finalMarking);
		}
		ws.workNet.setFinalMarkings(finalMarkings);
		// Update layout
		Transition bestTransition = null;
		Place bestPlace = null;
		if (transitions.size() == 1 && ws.postsets.get(transitions.iterator().next()).size() == 1) {
			bestTransition = transitions.iterator().next();
		} else if (places.size() == 1 && ws.presets.get(places.iterator().next()).size() == 1) {
			bestPlace = places.iterator().next();
		} else {
			int bestValue = Integer.MAX_VALUE;
//			for (Transition transition : transitions) {
//				if (ws.postsets.get(transition).size() < bestValue) {
//					bestValue = ws.postsets.get(transition).size();
//					bestTransition = transition;
//				}
//			}
			for (Place place : places) {
				if (ws.presets.get(place).size() < bestValue) {
					bestValue = ws.presets.get(place).size();
					bestTransition = null;
					bestPlace = place;
				}
			}
		}
//		ws.reverseTransitionMap.get(transition).getAttributeMap().put(AttributeMap.EDGECOLOR, Color.RED);
//		ws.reversePlaceMap.get(place).getAttributeMap().put(AttributeMap.STROKECOLOR, Color.RED);
		if (bestTransition != null) {
			Block block = new HorizontalBlock(ws.layoutMap.get(bestTransition),
					ws.layoutMap.get(place), ws.layoutMap.get(transition), ws);
			ws.layoutMap.put(bestTransition, block);
			for (Transition transition : transitions) {
				if (transition != bestTransition) {
					ws.hardTransitions.add(transition);
				}
			}

		} else if (bestPlace != null) {
			Block block = new HorizontalBlock(ws.layoutMap.get(place),
					ws.layoutMap.get(transition), ws.layoutMap.get(bestPlace), ws);
			ws.layoutMap.put(bestPlace, block);
			for (PetrinetNode node : ws.presets.get(bestPlace)) {
				if (node != transition) {
					ws.hardTransitions.add((Transition) node);
				}
			}
		}
		ws.update();
	}

	private void drain(Marking marking, Place place) {
		while (marking.occurrences(place) > 0) {
			marking.remove(place);
		}
	}

	public String toString() {
		return "Abstraction(" + transitions + "," + place + "," + transition + "," + places + ")";
	}

	public double score(ReductionWorkspace ws) {
		if (transitions.isEmpty()) {
			return 0.0;
		}
//		System.out.println("[AbstractionReduction] Preset = " + ws.presets.get(place).size() +", transitions = " + transitions.size());
//		return 0.99 / ws.presets.get(place).size(); 
		return 0.99 / transitions.size();
//		return 0.99 / ws.presets.get(places.iterator().next()).size(); 
	}

	public boolean hasOptimalScore(ReductionWorkspace ws) {
		return score(ws) == 0.99;
	}
}
